#include <aio.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <strings.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>

#include "../params.h"
#include "../include/debug.h"
#include "ringbuffer.h"
#include "pipe.h"

#define N_PARALLEL_WRITES 1024
#define N_PARALLEL_READS 2

#define N_SEQ_OPTS 1

template <typename T_ale, typename T_part>
class AleDevice
{
    private:
        int fd;
        RingBuffer<uint32_t> *empty_ales;

    public:

        AleDevice(const char* device, const size_t device_bytes)
        {
            this->fd = open(device, O_LARGEFILE | O_RDWR | O_DIRECT);
            if (this->fd < 0) 
            {
                INFO_DEBUG(1, "device: %s\n", device);
                perror("error in open");
                exit(1);
            }

            size_t nales = device_bytes / sizeof(T_ale);

            if (access(ALE_PATH "empty_ales.bin", F_OK) == 0)
            {
                // file exists
                empty_ales = new RingBuffer<uint32_t>(nales, ALE_PATH "empty_ales.bin");
            }
            else
            {
                // file did not exist, so we must add initial data
               empty_ales = new RingBuffer<uint32_t>(nales, ALE_PATH "empty_ales.bin");

                INFO_DEBUG(VERBOSITY > 0, "File 'empty_ales.bin' did not exist; putting values...\n");

                 for (uint32_t i = 0; i < nales; i++)
                    empty_ales->put(i);
            }
        }

        ~AleDevice()
        {
            delete this->empty_ales;
        }

        uint32_t alloc_ale()
        {
            ERROR_DEBUG(this->empty_ales->is_empty(), NOHALT, "There are no empty ales left!\n");

            return this->empty_ales->get();
        }

        void free_ale(uint32_t pos)
        {
            this->empty_ales->put(pos);
        }

        int write_ale(uint32_t pos, T_ale *data)
        {
            ssize_t rw_err;

            rw_err = pwrite(this->fd, (void*)data, sizeof(T_ale), (off64_t)pos * sizeof(T_ale));

            if(rw_err != sizeof(T_ale))
            {
                perror("pwrite");
                exit(rw_err);
            }

            return 0;
        }

        uint64_t read_ale(uint32_t pos, T_ale *data)
        {
            ssize_t rw_err;

            rw_err = pread(this->fd, (void*)data, sizeof(T_ale), (off64_t)pos * sizeof(T_ale));

            if(rw_err != sizeof(T_ale))
            {
                perror("pread");
                exit(rw_err);
            }

            return 0;
        }   
};

template <typename T_ale, typename T_part>
class AleSystem
{
    private:
        AleDevice<T_ale, T_part> *device;
 
        struct alepart
        {
            uint32_t *ale;
            uint32_t size;
            uint32_t nelems; //TODO: fix back to 64!
        };

        struct alepart *parts;
        size_t nparts;

        bool persist;

        int fd;
        uint8_t *data;


        void init(const size_t nparts, AleDevice<T_ale, T_part> *device, bool persist, uint8_t level, uint8_t list)
        {
            this->device = device;
            this->nparts = nparts;
            this->persist = persist;

//            if (persist)
            {
                char *fname;
                struct stat st;
                int r;

                fname = (char*)malloc(sizeof(ALE_PATH "ale_data_X_XX.bin"));

                snprintf(fname, sizeof(ALE_PATH "ale_data_X_XX.bin"), ALE_PATH "ale_data_%u_%02u.bin", level, list);

                fd = open(fname, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

                fstat(fd, &st);

                if (st.st_size == 0) // file did not exist before open
                {
                    INFO_DEBUG(VERBOSITY > 0, "File '%s' did not exist; resizing to %lu\n", fname,
                            (4*1024 + ((nparts * sizeof(struct alepart) + 
                                    nparts * (sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t)) + 
                                4*1024))  & ~(4*1024 - 1));

                    r = ftruncate(fd, (4*1024 + ((nparts * sizeof(struct alepart) + 
                                    nparts * (sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t)) + 
                                    4*1024))  & ~(4*1024 - 1));
                    //XXX-TODO: Do something with r
                }
                else
                {
                    lseek(fd, 0, SEEK_SET);

                    r = read(fd, (void*)&this->persist, sizeof(this->persist));
                    //XXX-TODO: Do something with r

                    INFO_DEBUG(VERBOSITY > 0, "File '%s' existed before; persist: %s\n", 
                            fname, this->persist ? "true" : "false");
                }

//                printf("opening %s\n", fname);

                data = (uint8_t*)mmap((caddr_t)0, 
                        ((nparts * sizeof(struct alepart) + nparts * (sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t))
                            + 4*1024)  & ~(4*1024 - 1), 
                        PROT_READ | PROT_WRITE, MAP_SHARED, fd, 4*1024);

//                printf("size: %u\n", nparts * sizeof(struct alepart) + nparts * (sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t));

                if (data == MAP_FAILED)
                    perror("error on mmap");

                data[0] = 0;

                this->parts = (struct alepart *)data;

                for (uint32_t j = 0; j < nparts; j++)
                {
                    (parts + j)->ale = (uint32_t*)(data + nparts * sizeof(struct alepart) + 
                            j * ((sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t)));
                }            
            }
/*            else
            {
                this->parts = (struct alepart *)malloc(nparts * sizeof(struct alepart));
                if(this->parts == NULL) 
                    perror("malloc");

                for (uint32_t j = 0; j < nparts; j++)
                {
                    (parts + j)->size  = 0;
                    (parts + j)->nelems = 0;

                    (parts + j)->ale   = (uint32_t*)malloc((sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t));
                    if((parts + j)->ale == NULL) 
                        perror("malloc");
                }
            }*/
        }



    public:

  /*      AleSystem(const size_t nparts, AleDevice<T_ale, T_part> *device)
        {
            init(nparts, device, false, 0, 0);
        }*/

        AleSystem(const size_t nparts, AleDevice<T_ale, T_part> *device, bool persist, uint8_t level, uint8_t list)
        {
            init(nparts, device, persist, level, list);
        }

        ~AleSystem()
        {
//            if (persist)
//            {
                lseek(fd, 0, SEEK_SET);

                int r;
                r = write(fd, (void*)&this->persist, sizeof(this->persist));
                //XXX-TODO: Do something with r

                munmap(data, 
                        ((nparts * sizeof(struct alepart) + nparts * (sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t))
                            + 4*1024)  & ~(4*1024 - 1));



                close(fd);
/*            }
            else
            {
                for (uint32_t j = 0; j < nparts; j++)
                {
                    free((parts + j)->ale);
                }

                free(parts);
            }*/
        }

        uint64_t get_list_size()
        {
            uint64_t ret = 0;
            uint64_t nelems;
               
            for(uint32_t i=0; i < this->nparts; i++)
            {
                nelems = this->parts[i].nelems;

                INFO_DEBUG(VERBOSITY > 2, "Size of part %u: %lu\n", i, nelems);

                ret += nelems;
            }
            return ret;
        }


        int write_ale(uint32_t part, uint64_t nelems, T_ale *data)
        {
            int ret;

            INFO_DEBUG(VERBOSITY > 5, "Writing block to part %u; nelems: %lu\n", part, nelems);

            ERROR_DEBUG(part > this->nparts, HALT, "Part %u > %lu!\n", part, this->nparts);

            struct alepart *p;
            p = &(this->parts[part]);

            if (p->size + 1 > (sizeof(T_part)/sizeof(T_ale)))
            {
                INFO_DEBUG(1, "Overflow in writing ale to part! pos: %u max_pos: %lu  part: %u\n", p->size,  (sizeof(T_part)/sizeof(T_ale)), part);
                ret = -1;
            }
            else
            {
                struct alepart *p;
                uint32_t pos;

                ERROR_DEBUG(part > this->nparts, HALT, "Part %u > %lu!\n", part, this->nparts);

                p = &(this->parts[part]);
                pos =__sync_fetch_and_add(&(p->size), 1);
                INFO_DEBUG(VERBOSITY > 5, "Writing ale %u, part %u\n", pos, part);
                if(pos >= (sizeof(T_part)/sizeof(T_ale)))
                {
                    fprintf(stderr, "Overflow in writing ale to part! pos: %u max_pos: %lu  part: %u\n", pos,  (sizeof(T_part)/sizeof(T_ale)), part);
                    fprintf(stderr, "Total number of entries in list so far: %lu\n", this->get_list_size());
                    __sync_fetch_and_add(&(p->size), -1);
                }
                else
                {

                    p->ale[pos] = this->device->alloc_ale();

                    INFO_DEBUG(VERBOSITY > 9, "hdd id: %u\n", p->ale[pos]);

                    this->device->write_ale(p->ale[pos], data);

                    __sync_fetch_and_add(&(p->nelems), nelems);
                }
            }

            return 0;
        }

        uint64_t read_part(uint32_t part, T_part *data)
        {
            struct alepart *p;
            uint64_t ret;

            INFO_DEBUG(VERBOSITY > 5,"Reading part %u\n", part);

            ERROR_DEBUG(((this->parts[part]).size * sizeof(T_ale)) > sizeof(T_part), HALT, 
                    "Part has too many ales to fit to data buffer!\n  number of ales: %lu\n  size of ale: %lu\n  size of part: %lu\n", d, sizeof(T_ale), sizeof(T_part));

            p = &(this->parts[part]);

            for (uint32_t i = 0; i < p->size; i++)
            {

                INFO_DEBUG(VERBOSITY > 9, "hdd id: %u  data: %p\n", p->ale[i], 
                        ((T_ale *)data) + i);

                device->read_ale(p->ale[i], (T_ale *)data + i);

                if (!persist)
                    this->device->free_ale(p->ale[i]);
            }

            ret = p->nelems;

            if (!persist)
            {
                p->size = 0;
                p->nelems = 0;
            }

            return ret;
        }
};

