#ifndef GATHERSTREAM_H
#define GATHERSTREAM_H

#include "debug.h"
#include "ringbuffer.h"

template <typename T>
class DataStream
{
    public:
        class ProducerQueue;
        class ConsumerQueue;

        typedef void (*produce_func)(uint32_t id, ProducerQueue *queue);
        typedef void (*consume_func)(uint32_t id, ConsumerQueue *queue);

        typedef class ProducerQueue
        {
            private:
//                T *wrapper;

                RingBuffer<T*> *ret_queue;
                RingBuffer<T*> **in_queues;

                uint32_t id;

                uint32_t num_queues;

            protected:
                friend class DataStream<T>;

                produce_func produce;

                ProducerQueue(uint32_t id,  produce_func produce, 
                        RingBuffer<T*> *ret_queue, RingBuffer<T*> **in_queues, uint32_t num_queues)
                {
                    this->id = id;
                    this->produce = produce;
                    this->ret_queue = ret_queue;
                    this->in_queues = in_queues;
                    this->num_queues = num_queues;
                };

            public:
                void put(uint32_t queue, T *value)
                {
//                    ERROR_DEBUG(wrapper->value != 0, HALT, "Buffer-error! Buffer is not empty on producer put!\n")

//                    wrapper->producer_id = this->id;
//                    wrapper->value = value;

                    if (queue >= this->num_queues)
                        fprintf(stderr, "queue: %u is larger than max queue (%u) - id: %u\n", queue, this->num_queues, this->id);

                    in_queues[queue]->put(value);
                }

                T* get()
                {
  /*                  T* ret;

                    wrapper = ret_queues[this->id]->get();

                    ret = wrapper->value;

#ifdef DEBUG
                    wrapper->value = 0;
#endif
*/
                    return ret_queue->get();
                };

                void unget(T *value)
                {
                    ret_queue->put(value);
                }

                bool is_empty()
                {
                    return ret_queue->is_empty();
                }

        } ProducerQueue_t;

        typedef class ConsumerQueue
        {
            private:
    //            T *ret_wrapper;

                RingBuffer<T*> *ret_queue;
                RingBuffer<T*> **in_queues;

                uint32_t id;

            protected:
                friend class DataStream<T>;

                consume_func consume;

                ConsumerQueue(uint32_t id,  consume_func consume, 
                        RingBuffer<T*> *ret_queue, RingBuffer<T*> **in_queues)
                {
                    this->id = id;
                    this->consume = consume;
                    this->ret_queue = ret_queue;
                    this->in_queues = in_queues;
                }

            public:
                void put(T *value)
                {
                    this->ret_queue->put(value);
                }

                T* get()
                {
                    return this->in_queues[this->id]->get();
                }

                bool get(T** ret, uint32_t secs, uint32_t nsecs)
                {
                    return this->in_queues[this->id]->get(ret, secs, nsecs);
                }

                bool is_empty()
                {
                    return this->in_queues[this->id]->is_empty();
                }

        } ConsumerQueue_t;

    private:
        RingBuffer<T*> *ret_queue;
        RingBuffer<T*> **in_queues;

        uint32_t num_producer_threads;
        uint32_t num_consumer_threads;
        uint32_t num_producer_buffers;
        uint32_t num_consumer_buffers;

//        T **values;

        pthread_t *producer_threads;
        pthread_t *consumer_threads;

        produce_func produce;
        consume_func consume;

        T* init_data;


    public:
        DataStream(uint32_t num_producer_threads, uint32_t num_consumer_threads,
                uint32_t num_producer_buffers, uint32_t num_consumer_buffers, 
                produce_func produce, consume_func consume)
        {
            uint32_t i, j;

            this->num_producer_threads = num_producer_threads;
            this->num_consumer_threads = num_consumer_threads;
            this->num_producer_buffers = num_producer_buffers;
            this->num_consumer_buffers = num_consumer_buffers;

            this->produce = produce;
            this->consume = consume;

            init_data = (T*)mmap((void *) 0, num_producer_threads * num_producer_buffers * sizeof(T), 
                            PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, (int) -1, (off_t) 0);

            if (init_data == MAP_FAILED)
            {
                fprintf(stderr, "Mapping failed in DataStream!\n");
                exit(-1);
            }

//                      (T*)malloc(num_producer_threads * num_producer_buffers * sizeof(T));

//            ret_queue = (RingBuffer<T*>**)
//                malloc(sizeof(RingBuffer<T*>*) * num_producer_threads);

//            values = (T**)malloc(
//                    sizeof(T*) * num_producer_threads *
//                    num_producer_buffers);

            ret_queue = new RingBuffer<T*>(num_producer_buffers * num_producer_threads);

            for (i = 0; i < num_producer_threads; i++)
            {
//                ret_queues[i] = new RingBuffer<T*>(num_producer_buffers);
                for (j = 0; j < num_producer_buffers; j++)
                {
//                    this->values[i*num_producer_buffers+j] = new T();
//                    this->values[i*num_producer_buffers+j]->value = 
//                        &init_data[i*num_producer_buffers+j];
                    ret_queue->put(&init_data[i*num_producer_buffers+j]);
                }
            }

            in_queues = (RingBuffer<T*>**)
                malloc(sizeof(RingBuffer<T*>*) * num_consumer_threads);

            for (i = 0; i < num_consumer_threads; i++)
            {
                in_queues[i] = new RingBuffer<T*>(num_consumer_buffers);
            }

            producer_threads = (pthread_t*)malloc(num_producer_threads * 
                    sizeof(pthread_t));

            consumer_threads = (pthread_t*)malloc(num_consumer_threads * 
                    sizeof(pthread_t));
        }

        void start()
        {
            uint32_t i;
            for(i=0;i<num_producer_threads;i++)
            {
                pthread_create(&producer_threads[i], NULL, produce_gather, 
                        (void*)new ProducerQueue(i, produce, ret_queue, in_queues, num_consumer_threads));
            }

            for(i=0;i<num_consumer_threads;i++)
            {
                pthread_create(&consumer_threads[i], NULL, consume_gather, 
                        (void*)new ConsumerQueue(i, consume, ret_queue, in_queues));
            }
        }

        void join()
        {
            uint32_t i;
            for(i=0;i<num_producer_threads;i++)
            {
                pthread_join(producer_threads[i], NULL);
            }

            for(i=0;i<num_consumer_threads;i++)
            {
                pthread_join(consumer_threads[i], NULL);
            }
        }

        ~DataStream()
        {
            uint32_t i;//, j;

/*            for (i = 0; i < num_producer_threads; i++)
            {
                for (j = 0; j < num_producer_buffers; j++)
                    delete values[num_producer_buffers*i+j];
            }
 */           
            delete ret_queue;
            
//            free(values);
            //free(ret_queues);

            for (i = 0; i < num_consumer_threads; i++)
            {
                delete in_queues[i];
            }

            free(in_queues);

            free(consumer_threads);
            free(producer_threads);

            munmap(init_data, num_producer_threads * num_producer_buffers * sizeof(T));
        }

    private:

        static void *produce_gather(void* arg)
        {
            ProducerQueue *producer_queue = (ProducerQueue*)arg;

            producer_queue->produce(producer_queue->id,
                    producer_queue);

            delete producer_queue;

            return NULL;
        } 

        static void *consume_gather(void *arg)
        {
            ConsumerQueue *consumer_queue = (ConsumerQueue*)arg;

            consumer_queue->consume(consumer_queue->id,
                    consumer_queue);

            delete consumer_queue;

            return NULL;
        }
};

template <typename T>
class GatherStream : public DataStream<T>
{
    public:
        typedef class ProducerQueue : public DataStream<T>::ProducerQueue
    {
        public:
            void put(T *value)
            {
                DataStream<T>::ProducerQueue::put(0, value);
            }
    } ProducerQueue_t;

        typedef void (*produce_func)(uint32_t id, ProducerQueue *queue);

        GatherStream(uint32_t num_producer_threads, 
                produce_func produce,
                typename DataStream<T>::consume_func consume)
            : DataStream<T>(num_producer_threads, 1, 2, num_producer_threads,
                    (typename DataStream<T>::produce_func)produce, 
                    consume)
    {
    }
};

template <typename T>
class ScatterStream : public DataStream<T>
{
    public:
         ScatterStream(uint32_t num_consumer_threads, 
                uint32_t num_producer_buffers, uint32_t num_consumer_buffers, 
                 typename DataStream<T>::produce_func produce,
                typename DataStream<T>::consume_func consume)
            : DataStream<T>(1, num_consumer_threads, num_producer_buffers, num_consumer_buffers, 
                    produce, consume)
    {
        ;
    }
};


#endif

