#ifndef RINGBUFFER_H 
#define RINGBUFFER_H

#include <semaphore.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <pthread.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "pipe.h"

#define die(msg) { perror(msg); exit(EXIT_FAILURE); }

template <typename T>
class RingBuffer : public Pipe<T>
{
    private:
        T *values;

        volatile uint32_t get_nextfree;
        volatile uint32_t get_firstoccupied;
        volatile uint32_t put_nextfree;
        volatile uint32_t put_firstoccupied;

        uint32_t bufsize;

        sem_t sem_get;
        sem_t sem_put;

        int fd;

        bool persist;

    public:
        RingBuffer(uint32_t bufnum, const char *fname = NULL)
        {
            int sem_get_val, sem_put_val;
            int r;

            // first init data incase file fname does not exist
            this->bufsize = bufnum + 1;
            this->get_nextfree = 0;
            this->get_firstoccupied = 0;
            this->put_nextfree = 0;
            this->put_firstoccupied = 0;

            sem_get_val = 0;
            sem_put_val = bufnum;

            persist = false;

            if (fname != NULL)
            {
                struct stat st;

                persist = true;

                fd = open(fname, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

                if (fd == -1)
                {
                    fprintf(stderr, "On opening file %s:\n", fname);
                    perror("");
                    exit(-1);
                }

                fstat(fd, &st);

                if (st.st_size == 0) // file did not exist before open; just adjust size but dont load data
                {
                    INFO_DEBUG(VERBOSITY > 0, "File '%s' did not exist; resizing to %lu\n", 
                            fname, (4*1024 + (sizeof(T) * bufsize + 4*1024)) & ~(4*1024 - 1));
                    r = ftruncate(fd, (4*1024 + (sizeof(T) * bufsize + 4*1024)) & ~(4*1024 - 1));
                    //XXX-TODO: Do something with r
                }
                else
                {
                    lseek(fd, 0, SEEK_SET);

                    r = read(fd, (void*)&this->get_nextfree, sizeof(this->get_nextfree));
                    //XXX-TODO: Do something with r
                    r = read(fd, (void*)&this->get_firstoccupied, sizeof(this->get_firstoccupied));
                    //XXX-TODO: Do something with r
                    r = read(fd, (void*)&this->put_nextfree, sizeof(this->put_nextfree));
                    //XXX-TODO: Do something with r
                    r = read(fd, (void*)&this->put_firstoccupied, sizeof(this->put_firstoccupied));
                    //XXX-TODO: Do something with r
                    r = read(fd, (void*)&sem_get_val, sizeof(sem_get_val));
                    //XXX-TODO: Do something with r
                    r = read(fd, (void*)&sem_put_val, sizeof(sem_put_val));
                    //XXX-TODO: Do something with r

                    INFO_DEBUG(VERBOSITY > 0, "File '%s' allready existet...\n", fname);
                }

                values = (T*)mmap((caddr_t)0, (sizeof(T) * bufsize + 4*1024) & ~(4*1024 - 1),
//                        nparts * sizeof(struct alepart) + nparts * (sizeof(T_part)/sizeof(T_ale)) * sizeof(uint32_t), 
                        PROT_READ | PROT_WRITE, MAP_SHARED, fd, 4*1024);

                if (values == MAP_FAILED)
                {
                    perror("mmap did not succeed");
                    exit(-1);
                }

            }
            else
            {
                values = (T*)malloc(sizeof(T) * bufsize);
            }

            if (sem_init(&sem_get, 0, sem_get_val) != 0)
            {
                fprintf(stderr, "sem_get_val: %u\n", sem_get_val);
                perror("sem_init failed in RingBuffer");
                exit(-1);
            }
            if (sem_init(&sem_put, 0, sem_put_val) != 0)
            {
                fprintf(stderr, "sem_put_val: %u  bufnum: %u\n", sem_put_val, bufnum);
                perror("sem_init failed in RingBuffer");
                exit(-1);
            }

//            printf("sem_get_val: %u, sem_put_val: %u\n", sem_get_val, sem_put_val);
         }

        ~RingBuffer()
        {
            if (persist)
            {
                int sem_get_val, sem_put_val;
                int r;

                if (sem_getvalue(&sem_get, &sem_get_val) != 0)
                    perror("can't read sem value");
                if (sem_getvalue(&sem_put, &sem_put_val) != 0)
                    perror("can't read sem value");


                lseek(fd, 0, SEEK_SET);

                r = write(fd, (void*)&this->get_nextfree, sizeof(this->get_nextfree));
                //XXX-TODO: Do something with r
                r = write(fd, (void*)&this->get_firstoccupied, sizeof(this->get_firstoccupied));
                //XXX-TODO: Do something with r
                r = write(fd, (void*)&this->put_nextfree, sizeof(this->put_nextfree));
                //XXX-TODO: Do something with r
                r = write(fd, (void*)&this->put_firstoccupied, sizeof(this->put_firstoccupied));
                //XXX-TODO: Do something with r
                r = write(fd, (void*)&sem_get_val, sizeof(sem_get_val));
                //XXX-TODO: Do something with r
                r = write(fd, (void*)&sem_put_val, sizeof(sem_put_val));
                //XXX-TODO: Do something with r

//               printf("%u %u %u %u %u %u\n", this->get_nextfree, this->get_firstoccupied, 
//                       this->put_nextfree, this->put_firstoccupied, sem_get_val, sem_put_val);


                munmap(values, (sizeof(T) * bufsize + 4*1024)  & ~(4*1024 - 1));

                close(fd);
            }
            else
            {
                free(values);
            }

            sem_destroy(&sem_get);
            sem_destroy(&sem_put);
        }

        inline uint32_t get_bufnum()
        {
            return this->bufsize - 1;
        }

        inline bool is_full()
        {
            return ((this->put_nextfree + 1) % bufsize) == (this->put_firstoccupied % bufsize);
        }

        inline bool is_empty()
        {
            return (this->get_firstoccupied % bufsize) == (this->get_nextfree % bufsize);
        }

        T get()
        {
            T ret;

            sem_wait(&sem_get);

            ret = do_get();

            sem_post(&sem_put);

            return ret;
        }

        void unget(T buf)
        {
            put(buf);
        }

        void put(T buf) 
        {
            sem_wait(&sem_put);

            do_put(buf);

            sem_post(&sem_get);
        }

        bool get(T *buf, uint32_t secs, uint32_t nsecs)
        {
            struct timespec tp;

            int s;

            if (clock_gettime(CLOCK_REALTIME, &tp) == -1)
                die("clock_gettime");

            tp.tv_sec += secs;
            tp.tv_nsec += nsecs;
            
            while ((s = sem_timedwait(&sem_get, &tp)) == -1 && errno == EINTR)
                continue;       /* Restart when interrupted by handler */

            if (s == -1) 
                if (errno == ETIMEDOUT)
                    return false;

            *buf = do_get();

            sem_post(&sem_put);

            return true;
        }

        bool put(T buf, uint32_t secs, uint32_t nsecs)
        {
            struct timespec tp;

            int s;

            if (clock_gettime(CLOCK_REALTIME, &tp) == -1)
                die("clock_gettime");

            tp.tv_sec += secs;
            tp.tv_nsec += nsecs;

            while ((s = sem_timedwait(&sem_put, &tp)) == -1 && errno == EINTR)
                continue;       /* Restart when interrupted by handler */

            if (s == -1) 
                if (errno == ETIMEDOUT)
                    return false;

            do_put(buf);

            sem_post(&sem_get);

            return true;
        }


    private:

        inline T do_get()
        {
            uint32_t loc_firstoccupied;
            T ret;

            do
            {
                loc_firstoccupied = this->get_firstoccupied;

                while (((loc_firstoccupied) % bufsize) == (this->get_nextfree % bufsize))
                    sched_yield();
            } 
            while (!__sync_bool_compare_and_swap(&this->get_firstoccupied, loc_firstoccupied,
                        (loc_firstoccupied + 1) % bufsize));

            ret = values[(loc_firstoccupied) % bufsize];

            while (!__sync_bool_compare_and_swap(&this->put_firstoccupied, loc_firstoccupied,
                        (loc_firstoccupied + 1) % bufsize))
                sched_yield();

            return ret;
        }

        inline void do_put(T buf)
        {
            uint32_t loc_nextfree;


            do
            {
                loc_nextfree = this->put_nextfree;

                while (((loc_nextfree + 1) % bufsize) == (this->put_firstoccupied % bufsize))
                    sched_yield();
            }
            while (!__sync_bool_compare_and_swap(&this->put_nextfree, loc_nextfree, 
                        (loc_nextfree + 1) % bufsize));

            values[loc_nextfree % bufsize] = buf;

            while (!__sync_bool_compare_and_swap(&this->get_nextfree, loc_nextfree, 
                        (loc_nextfree + 1) % bufsize))
                sched_yield();
        }

};

#endif

