#ifndef TASK_H
#define TASK_H

#include <stdlib.h>

#include "pipe.h"

class Task
{
    private:
        int tid;

    public:
        Task()
        {
            tid = -1;
        }

        Task(int tid)
        {
            this->tid = tid;
        }

        virtual void run(void* argv, int tid) = 0;

        virtual void run(void *argv)
        {
           this->run(argv, this->tid);
        }

        inline int getTID()
        {
            return this->tid;
        }

        void setTID(int tid)
        {
            this->tid = tid;
        }
};

template <typename SourceT, typename SinkT>
class BasicTask : public Task
{
    private:
        SinkT *sink;
        SourceT *source;

    public:

        SourceT *getSource()
        {
            return source;
        }

        SinkT *getSink()
        {
            return sink;
        }

        void bind(SourceT *source, SinkT *sink)
        {
            this->sink = sink;
            this->source = source;
        }

};

template <typename SourceA, typename SinkA, typename SourceB, typename SinkB>
class BasicCrossTask : public Task
{
    private:
        SourceA *sourceA;
        SinkA   *sinkA;

        SourceB *sourceB;
        SinkB   *sinkB;

    public:
        SourceB *getSourceB()
        {
            return sourceB;
        }

        SinkB *getSinkB()
        {
            return sinkB;
        }

        SourceA *getSourceA()
        {
            return sourceA;
        }

        SinkA *getSinkA()
        {
            return sinkA;
        }

        void bind(SourceA *sourceA, SinkA *sinkA, SourceB *sourceB, SinkB *sinkB)
        {
            this->sinkA = sinkA;
            this->sourceA = sourceA;
            this->sinkB = sinkB;
            this->sourceB = sourceB;
        }
};

template <typename T>
class PipelineTask : public BasicTask<Source<T>, Sink<T> > { };
 
template <typename T>
class ScatterTask : public BasicTask<Source<T>, SinkList<T> > { };

template <typename T, typename U>
class CrossTask : public BasicCrossTask<SourceList<T>, SinkList<T>, Source<U>, SinkList<U> > { };

class Tasks
{
    protected:
        int num_tasks;

    public:
        virtual Task *getTask(int tid) = 0;

        int getNumTasks()
        {
            return num_tasks;
        }
};

#define TASK_ARG(task) task ## Arg

template <typename T>
class TaskGroup : public Tasks
{
    protected:
        Task **task;

    public:
        Task *getTask(int tid)
        {
            return task[tid];
        }
        
        template <typename U, typename V>
        void bind(SourceList<U> *sources, Sink<V> *sink)
        {
            for (int i = 0; i < num_tasks; i++)
            {
                ((T*)(task[i]))->bind(sources->getSource(i), sink);
            }
        }

        template <typename U, typename V>
        void bind(Source<U> *source, SinkList<V> *sinks)
        {
            for (int i = 0; i < num_tasks; i++)
            {
                ((T*)(task[i]))->bind(source, sinks);
            }
        }

        template <typename U, typename V>
        void bind(SourceList<U> *sources, SinkList<V> *sinks)
        {
            for (int i = 0; i < num_tasks; i++)
            {
                ((T*)(task[i]))->bind(sources->getSource(i), sinks->getSink(i));
            }
        }

        template <typename U, typename V>
        void bind(SourceList<U> *sourcesA, Sink<U> *sinkA, Source<V> *sourceB, Sink<V> *sinkB)
        {
            for (int i = 0; i < num_tasks; i++)
            {
                ((T*)(task[i]))->bind(sourcesA->getSource(i), sinkA, sourceB, sinkB);
            }
        }

        TaskGroup(int num_tasks)
        {
            this->num_tasks = num_tasks;

            this->task = (Task **)malloc(num_tasks * sizeof(Task *));

            for (int i = 0; i < num_tasks; i++)
            {
                task[i] = new T;
                ((T*)(task[i]))->setTID(i);
            }
        }

/*	TaskGroup(int num_tasks, TASK_ARG(T) arg)
	{
		this->num_tasks = num_tasks;

		this->task = (Task **)malloc(num_tasks * sizeof(Task *));

		for (int i = 0; i < num_tasks; i++)
		{
			task[i] = new T(arg);
			((T*)(task[i]))->setTID(i);
		}
	}*/

        ~TaskGroup()
        {
            for (int i = 0; i < num_tasks; i++)
            {
                delete task[i];
            }

            free(this->task);
        }
};


#endif

