#include <stdio.h>
#include <mpi.h>

#include <getopt.h>
#include <ctype.h>
#include <assert.h>

#include <sched.h>
#include <sys/types.h>
#include <linux/unistd.h>
#include <errno.h>
#include <pthread.h>

#include "types.h"
#include "argv.h"
#include "netdata.h"
#include "datastream/alesystem.h"
#include "datastream/buffermanager.h"
#include "datastream/ringbuffer.h"
#include "params.h"
#include "sigsegv.h"

#include "pipe.h"
#include "task.h"
#include "thread.h"


struct timeval start, end;

AleDevice<ale_t, part_t> *aledev;

int numprocs, rank, success;

volatile int running = 1;

#include "store.h"
#include "presort.h"
#include "sort.h"
#include "load.h"
#include "merge.h"
#include "recv.h"
#include "send.h"

void show_help(void)
{
    printf("\nUsage: mpirun -l -machinefile hosts -np 8 ./fsbday [OPTIONS] [CLAMPLIST]\n"\
            "Options:\n\n"\
            "\t--nlists\n"\
            "\t--startlevel\n"\
            "\t--startlist\n"\
            "\t--persist\n"\
            "\t-h  this text\n\n");
}

int main(int argc, char *argv[])
{

    // INVARIANT CHECK
    assert((ALESIZE % sizeof(Entry<0>)) == 0);
    assert((ALESIZE % sizeof(Entry<1>)) == 0);
    assert((ALESIZE % sizeof(Entry<2>)) == 0);
    assert((ALESIZE % sizeof(Entry<3>)) == 0);
    assert((ALL_BITS % 8) == 0);

    char processor_name[MPI_MAX_PROCESSOR_NAME];
    int has_threads, namelen, i;

    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &has_threads);

    ERROR_DEBUG(has_threads != MPI_THREAD_MULTIPLE, HALT, 
            "MPI does not support MPI_THREAD_MULTIPLE!\n");

    Argv *arg = new Argv();

    success = 0;

    {
        uint8_t xors = 0;

        while (1)
        {
            int option_index = 0;
            static struct option long_options[] = {
                {"startlevel", required_argument, 0, 0},
                {"nlists", required_argument, 0, 0},
                {"startlist", required_argument, 0, 0},
                {"persist", no_argument, 0, 0},
                {0, 0, 0, 0}
            };

            char c = getopt_long (argc, argv, "h",
                    long_options, &option_index);
            if (c == -1)
                break;


            switch (c)
            {
                case 0:
                    if (option_index == 0) // startlevel
                    {
                        arg->startlevel = atoi(optarg);
                    }
                    if (option_index == 1) // nlists
                    {
                        arg->startnlists = atoi(optarg);
                    }
                    if (option_index == 2) // startlist
                    {
                        arg->startlist = atoi(optarg);
                    }
                    if (option_index == 3) // persist
                    {
                        arg->persist = true;
                    }
                    break;
                case 'h':
                    show_help();
                    exit(0);
                    break;
                case '?':
                    if (optopt == 'n')
                        fprintf (stderr, "Option -%c requires an argument.\n", optopt);
                    else if (isprint (optopt))
                        fprintf (stderr, "Unknown option `-%c'.\n", optopt);
                    else
                        fprintf (stderr,
                                "Unknown option character `\\x%x'.\n",
                                optopt);
                    return 1;
                default:
                    abort ();
            }
        }

        if (arg->startlevel < 0)
            arg->startlevel = 0;
        if (arg->startlevel > MAX_LEVEL)
            arg->startlevel = MAX_LEVEL;
        fprintf(stderr, "startlevel: %u\n", arg->startlevel);

        if (arg->startnlists < 2)
            arg->startnlists = 2;
        if (arg->startnlists > NNODES)
            arg->startnlists = NNODES;
        fprintf(stderr, "nlists: %u\n", arg->startnlists);

        if (arg->startlist < 0)
            arg->startlist = 0;
        if (arg->startlist + arg->startnlists > (1<<MAX_LEVEL))
            arg->startlist = (1<<MAX_LEVEL) - arg->startnlists;
        fprintf(stderr, "startlist: %u\n", arg->startlist);

        fprintf(stderr, "persist: %u\n", arg->persist);

        arg->clamp = (uint8_t*)malloc(arg->startnlists * sizeof(uint8_t));

        for (uint8_t i = 0; i < arg->startnlists; i++)
            arg->clamp[i] = 0;

        for (uint8_t i = 0; (i < arg->startnlists) && (optind + i < argc); i++)
        {
           arg->clamp[i] = atoi(argv[optind + i]);
           if (arg->clamp[i] > MAXCLAMP)
           {
               fprintf(stderr, "Clamp value no. %u exceeds maximum clamping value!", i);
               exit(0);
           }
        }

        fprintf(stderr, "clamp values: ");
        for (uint8_t i = 0; i < (int)arg->startnlists; i++)
        {
            fprintf(stderr, "%i ", arg->clamp[i]);
            xors ^= arg->clamp[i];
        }
        fprintf(stderr, "\n");

        if (xors != 0)
        {
            fprintf(stderr, "Clamp values do not xor to 0!\n Killing job...\n");
            exit(0);
        }
    }

    setup_sigsegv();

    if (mprotect(bytematrix, sizeof(bytematrix), PROT_READ) == -1)
    {
        perror("Protecting bytematrix");
    }

    gettimeofday(&start, NULL);

    MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Get_processor_name(processor_name, &namelen);


    fprintf(stderr, "I am %s\n", processor_name);

    MPI_Barrier(MPI_COMM_WORLD);

    aledev = new AleDevice<ale_t, part_t>(IODEVICE, DEVICE_BYTES);
    
    /** recv/store *
     **
     **  |------------------------|
     **  |                        |  |---------------------|
     **  |--> Recv >===|--|       ^  v                     |
     **  |--> Recv >===|--+---> Presort >---|              |
     **  |--> Recv >===|--|                 |              |
     **  |--> Recv >===|--|                 |---> Store >--|
     **                |  |                 |              |
     **       ... 8x   |------> Presort >---|              |
     **                |  |       v  ^                     |
     **  |--> Recv >===|--|       |  |---------------------|
     **  |                        |
     **  |------------------------|
     **/
    TaskGroup<Recv> recv(8);
    TaskGroup<Presort> presort(2);
    Store store;

    BufferManager<NetData> presort_recv(16);
    PipeList<NetData*> recv_presort(2);

    BufferManager<Ale> store_presort(2 * NPARTS);
    RingBuffer<Ale*> presort_store(2 * NPARTS);

    recv_presort.add(new RingBuffer<NetData*>(16));
    recv_presort.add(new RingBuffer<NetData*>(16));

    recv.bind(&presort_recv, &recv_presort);
    presort.bind(&recv_presort, &presort_recv, &store_presort, &presort_store);

    store.bind(&presort_store, &store_presort);

    ThreadGroup recv_threads(&recv);
    ThreadGroup presort_threads(&presort);
    Thread store_thread(&store);

    /** load/send *
     **       
     **  Load ----> Sort ----> |-------| ----> Send ---|
     **   ^                    |       | ----> Send ---|
     **   |------------------  |       | ----> Send ---|
     **                        | Merge | ----> Send ---|
     **   |------------------  |       | ... 8x        |
     **   v                    |       | ----> Send ---|
     **  Load ----> Sort ----> |-------| ----> Send ---|
     **                            ^                   |
     **                            |-------------------|                    
     **
     **/
    TaskGroup<Load> load(2);
    TaskGroup<Sort> sort(2);
    Merge merge;
    TaskGroup<Send> send(8);


    BufferManager<NetData> send_merge(16);
    PipeList<NetData*> merge_send(8);

    PipeList<Part*> merge_load(2);
    PipeList<Part*> load_sort(2);
    PipeList<Part*> sort_merge(2);

    merge_load.add(new BufferManager<Part>(3));
    merge_load.add(new BufferManager<Part>(3));

    load_sort.add(new RingBuffer<Part*>(3));
    load_sort.add(new RingBuffer<Part*>(3));

    sort_merge.add(new RingBuffer<Part*>(3));
    sort_merge.add(new RingBuffer<Part*>(3));

    for (i = 0; i < 8; i++)
        merge_send.add(new RingBuffer<NetData*>(2));    

    merge.bind(&sort_merge, &merge_load, &send_merge, &merge_send);
    send.bind(&merge_send, &send_merge);

    load.bind(&merge_load, &load_sort);
    sort.bind(&load_sort, &sort_merge);

    Thread merge_thread(&merge);
    ThreadGroup send_thread(&send);

    ThreadGroup load_thread(&load);
    ThreadGroup sort_thread(&sort);

    /** run... **/

    recv_threads.start(arg);
    presort_threads.start(arg);
    store_thread.start(arg);

    load_thread.start(arg);
    sort_thread.start(arg);
    merge_thread.start(arg);
    send_thread.start(arg);

    load_thread.join();
    sort_thread.join();
    merge_thread.join();
    presort_threads.join();
    store_thread.join();

    MPI_Barrier(MPI_COMM_WORLD);
    // Terminate send threads
    running = 0;

    // Shutdown blocking MPI_Recv in recv-thread by sending empty packet:
    for(i=0;i<NNODES;i++) 
        MPI_Send (NULL, 0, MPI_BYTE, i, 0, MPI_COMM_WORLD);

    recv_threads.join();
    send_thread.join();

    recv_presort.deleteContent<RingBuffer<NetData*> >();
    merge_send.deleteContent<RingBuffer<NetData*> >();

    delete aledev;

    MPI_Finalize();

    return success;
}

